﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Media.Imaging;
using System.Windows.Media;
using System.Windows;
using System.IO;
using System.Security.Permissions;
using System.Runtime.InteropServices;



namespace DrawingBoard
{
    /// <summary>
    /// Contains helper methods used when painting the image.
    /// </summary>
    public class BitmapHelper
    {
        #region private Data

        const int AirbrushDots = 32;
        const int AirbrushDotRadius = 9;
        private static byte[][] AirbrushDotKernel;

        private List<MovingBrushStroke> brushStrokes;

        private List<DrawingImage> baseBrushes;
        private List<DrawingImage> brushes;
        private DrawingImage backImage1;
        private DrawingImage circleImage1;
        #endregion

        #region Definitions

        /// <summary>
        /// Represents a bounding box, used here for determining drawing bounds
        /// </summary>
        private struct BoundingBox
        {
            #region Properties

            /// <summary>
            /// Gets whether the bounding box is bounding any content
            /// </summary>
            public bool HasContent { get; private set; }

            /// <summary>
            /// The left edge of the bounding box
            /// </summary>
            public int Left { get; private set; }

            /// <summary>
            /// The right edge of the bounding box
            /// </summary>
            public int Right { get; private set; }

            /// <summary>
            /// The top edge of the bounding box
            /// </summary>
            public int Top { get; private set; }

            /// <summary>
            /// The bottom edge of the bounding box
            /// </summary>
            public int Bottom { get; private set; }

            /// <summary>
            /// The width of the bounding box
            /// </summary>
            public int Width { get; private set; }

            /// <summary>
            /// The height of the bounding box
            /// </summary>
            public int Height { get; private set; }


            #endregion

            #region Methods

            /// <summary>
            /// Adds a point that the bounding box must contain
            /// </summary>
            /// <param name="x">X coordinate of the point</param>
            /// <param name="y">Y coordinate of the point</param>
            /// <param name="radius">Radius of the point</param>
            public void AddPoint(int x, int y, int radius)
            {
                if (!HasContent)
                {
                    Left = x - radius;
                    Right = x + radius + 1;
                    Top = y - radius;
                    Bottom = y + radius + 1;
                    Width = Height = 2 * radius;
                    HasContent = true;
                }
                else
                {
                    if (x - radius < Left)
                        Left = x - radius;
                    if (x + radius + 1 > Right)
                        Right = x + radius + 1;
                    if (y - radius < Top)
                        Top = y - radius;
                    if (y + radius + 1 > Bottom)
                        Bottom = y + radius + 1;
                    Width = Right - Left;
                    Height = Bottom - Top;
                }
            }

            /// <summary>
            /// Clips this bounding box against a larger bounding region
            /// </summary>
            /// <param name="clipregion">The bounding box of the clipping region</param>
            public void Clip(BoundingBox clipregion)
            {
                if (Left < clipregion.Left)
                    Left = clipregion.Left;
                if (Left > clipregion.Right)
                    Left = clipregion.Right;
                if (Right < clipregion.Left)
                    Right = clipregion.Left;
                if (Right > clipregion.Right)
                    Right = clipregion.Right;
                if (Top < clipregion.Top)
                    Top = clipregion.Top;
                if (Top > clipregion.Bottom)
                    Top = clipregion.Bottom;
                if (Bottom < clipregion.Top)
                    Bottom = clipregion.Top;
                if (Bottom > clipregion.Bottom)
                    Bottom = clipregion.Bottom;
                Width = Right - Left;
                Height = Bottom - Top;
            }

            #endregion
        }

        /// <summary>
        /// Represents a line segment, and optimizes the calculation of its nearest distance to multiple points
        /// </summary>
        private struct LineSegment
        {
            #region Data

            int fromx, fromy, tox, toy;     // Stores the 'to' and 'from' points as integers

            int bx, by, bxx, bxy, byy, L2;  // Stores intermediate calculations, so they aren't needlessly repeated

            #endregion

            /// <summary>
            /// Constructor
            /// </summary>
            /// <param name="_from">The starting point of the line segment</param>
            /// <param name="_to">The ending point of the line segment</param>
            public LineSegment(Point _from, Point _to)
            {
                fromx = (int)_from.X;
                fromy = (int)_from.Y;
                tox = (int)_to.X;
                toy = (int)_to.Y;
                bx = tox - fromx;
                by = toy - fromy;
                bxx = bx * bx;
                bxy = bx * by;
                byy = by * by;
                L2 = bxx + byy;
            }

            #region Methods

            /// <summary>
            /// Returns the squared distance between the line segment and the point represented by x,y
            /// </summary>
            /// <param name="x">X coordinate of the point</param>
            /// <param name="y">Y coordinate of the point</param>
            /// <returns></returns>
            public int DistanceSquared(int x, int y)
            {
                int nx, ny;

                int ay = y - fromy;
                int ax = x - fromx;

                // Compute the point along the line nearest to the current pixel (represented by nx,ny)                    
                if (L2 != 0)
                {
                    nx = (bxx * ax + bxy * ay) / L2;
                    ny = (bxy * ax + byy * ay) / L2;

                    // If the nearest point lies beyond the 'to' point, snap n to 'to'
                    if (nx * nx + ny * ny > L2)
                    {
                        nx = bx;
                        ny = by;
                    }
                    else
                    {
                        int bnx = bx - nx;
                        int bny = by - ny;

                        // If the nearest point lies before the 'from' point, snap n to 'from'
                        if (bnx * bnx + bny * bny > L2)
                            nx = ny = 0;
                    }
                }
                else
                    nx = ny = 0;

                // Return the squared distance between n and the given point
                int dx = nx - ax;
                int dy = ny - ay;

                return dx * dx + dy * dy;
            }

            /// <summary>
            /// Interpolates along the line segment
            /// </summary>
            /// <param name="i">The step index to interpolate to</param>
            /// <param name="steps">The number of steps</param>
            /// <param name="x">The returned x coordinate</param>
            /// <param name="y">The returned y coordinate</param>
            public void Interpolate(int i, int steps, out int x, out int y)
            {
                x = fromx + bx * i / steps;
                y = fromy + by * i / steps;
            }

            #endregion
        }

        #endregion

        #region Constructor

        public BitmapHelper()
        {
            brushStrokes = new List<MovingBrushStroke>();
            AirbrushDotKernel = new byte[2 * AirbrushDotRadius + 1][];

            for (int i = 0; i < 2 * AirbrushDotRadius + 1; i++)
                AirbrushDotKernel[i] = new byte[2 * AirbrushDotRadius + 1];

            int x, y;
            for (int i = 0; i < AirbrushDotKernel.Length; i++)
            {
                x = i - AirbrushDotRadius;
                for (int j = 0; j < AirbrushDotKernel.Length; j++)
                {
                    y = j - AirbrushDotRadius;
                    AirbrushDotKernel[i][j] = (byte)(128 - (int)(128.0 * Math.Min(1, Math.Sqrt(x * x + y * y) / (double)AirbrushDotRadius)));
                }
            }

            baseBrushes = new List<DrawingImage>();
            brushes = new List<DrawingImage>();
            foreach (var brush in Directory.GetFiles(Path.GetFullPath(@"Files/Brushes")))
            {
                baseBrushes.Add(new DrawingImage(brush));
                brushes.Add(new DrawingImage(brush));
            }
                
            backImage1 = new DrawingImage(@"Files/blue077.jpg");
            circleImage1 = new DrawingImage(@"Files/circle.jpg");
        }

        #endregion

        public void LoadNewBackImage(string path)
        {
            backImage1 = new DrawingImage(path);
        }

        /// <summary>
        /// Paints on a pbgra32 WriteableBitmap with a stylized airbrush
        /// </summary>
        /// <param name="bmp">The bitmap to modify</param>
        /// <param name="from">The starting point of the stroke</param>
        /// <param name="to">The end point of the stroke</param>
        /// <param name="color">The color of the stroke</param>
        /// <param name="size">The size of the stroke</param>
        public unsafe void Airbrush(WriteableBitmap bmp, Point from, Point to, Color color, int size)
        {
            Random r = new Random();

            if (bmp == null) return;

            bmp.Lock();

            // Create a line segment representation
            LineSegment segment = new LineSegment(from, to);

            // Get a bounding box for the painted area
            BoundingBox bitmapbounds = new BoundingBox();
            BoundingBox segmentbounds = new BoundingBox();

            bitmapbounds.AddPoint(0, 0, 0);
            bitmapbounds.AddPoint(bmp.PixelWidth - 1, bmp.PixelHeight - 1, 0);

            segmentbounds.AddPoint((int)from.X, (int)from.Y, size + AirbrushDotRadius);
            segmentbounds.AddPoint((int)to.X, (int)to.Y, size + AirbrushDotRadius);
            segmentbounds.Clip(bitmapbounds);

            UInt32* start = (UInt32*)bmp.BackBuffer.ToPointer();
            int stride = bmp.BackBufferStride / sizeof(UInt32);
            // Move from 'from' to 'to' along timestep intervals, with one dot painted per interval
            for (int i = 0; i < AirbrushDots; i++)
            {
                int x, y;
                segment.Interpolate(i, AirbrushDots, out x, out y);

                int dist = r.Next() % size;
                double angle = r.NextDouble() * 2 * Math.PI;

                double dx = Math.Cos(angle) * dist;
                double dy = Math.Sqrt(dist * dist - dx * dx);
                if (angle > Math.PI) dy = -dy;

                int bx = x + (int)dx;
                int by = y + (int)dy;

                BoundingBox dotbounds = new BoundingBox();

                dotbounds.AddPoint(bx, by, AirbrushDotRadius);
                dotbounds.Clip(bitmapbounds);

                for (int k = dotbounds.Top, row = 0; k < dotbounds.Bottom; k++, y++, row++)
                    for (int j = dotbounds.Left, col = 0; j < dotbounds.Right; j++, col++)
                        WriteAlphaBlended(start + stride * k + j, Color.FromArgb(AirbrushDotKernel[row][col], color.R, color.G, color.B));
            }

            bmp.AddDirtyRect(new Int32Rect(segmentbounds.Left, segmentbounds.Top, segmentbounds.Width, segmentbounds.Height));
            bmp.Unlock();
        }

        /// <summary>
        /// draw ellipse 
        /// </summary>
        public void Ellipse(WriteableBitmap bmp, Point from, Point to, Color color)
        {
            if (bmp == null) return;
            int width = (int)(from.X>to.X ? from.X - to.X: to.X - from.X) ;
            int height = (int)(from.Y>to.Y ? from.Y - to.Y: to.Y - from.Y) ;
            bmp.Lock();
            for (int i = 0; i < 5; i++)
            {
                bmp.DrawEllipseCentered((int)to.X, (int)to.Y, width+i, height+i, color);
            }
            bmp.Unlock();
        }

        public void SetDrawImageSize(int size)
        {

            for (int i = 0; i < baseBrushes.Count(); i++)
                brushes[i].WriteableBitmap = Resize(baseBrushes[i].WriteableBitmap, size, size);
        }
        /// <summary>
        /// Painting on bmp the ink brush. 
        /// </summary>
        public unsafe void DrawImage(WriteableBitmap bmp, Point to, Color color)
        {
            bmp.Lock();

            // Get a bounding box for the painted area
            BoundingBox bitmapbounds = new BoundingBox();
            BoundingBox segmentbounds = new BoundingBox();

            //add the size of bmp to the bounding box
            bitmapbounds.AddPoint(0, 0, 0);
            bitmapbounds.AddPoint(bmp.PixelWidth - 1, bmp.PixelHeight - 1, 0);

            UInt32* start = (UInt32*)bmp.BackBuffer.ToPointer();
            int stride = bmp.BackBufferStride / sizeof(UInt32);

            int toX = (int)to.X;
            int toY = (int)to.Y;

            //set a random ink image
            Random rnd = new Random();
            var inkImage = brushes[rnd.Next(brushes.Count)].WriteableBitmap;

            //take the ink image width and height
            int width = (int)inkImage.PixelWidth;
            int height = (int)(inkImage.PixelHeight);

            //add the bounderis of the painted location
            segmentbounds.AddPoint((int)to.X - width / 2, (int)to.Y - height / 2, 0);
            segmentbounds.AddPoint((int)to.X + width / 2 -1 , (int)to.Y + height / 2 -1, 0);
            segmentbounds.Clip(bitmapbounds);

            //add pixel by pixel the ink image to the bmp 
            for (int k = segmentbounds.Top, row = 0; k < segmentbounds.Bottom ; k++, row++)
                for (int j = segmentbounds.Left, col = 0; j < segmentbounds.Right; j++, col++)
                {
                    Byte colorAlpha = inkImage.GetPixel(col, row).A;
                    WriteAlphaBlended(start + stride * k + j, Color.FromArgb(colorAlpha, color.R, color.G, color.B));
                }
            bmp.AddDirtyRect(new Int32Rect(segmentbounds.Left, segmentbounds.Top, segmentbounds.Width, segmentbounds.Height));
            bmp.Unlock();       
        }

        /// <summary>
        /// 
        /// </summary>
        public void MovingBrush(WriteableBitmap bmp, Point from, Point to, Color color)
        {
            MovingBrushStroke newBrushStroke = new MovingBrushStroke(from, to, color);
            brushStrokes.Add(newBrushStroke);
            foreach (var brushStroke in brushStrokes)
            {
                brushStroke.DrawNext(bmp);
            }
            //clear died brushStroke ( TimeToLive == 0)
            brushStrokes = brushStrokes.Where(brushStroke => (brushStroke.TimeToLive > 0)).ToList();
        }

        /// <summary>
        /// Painting on bmp the texture image. 
        /// </summary>
        public unsafe void DrawBackImage(WriteableBitmap bmp, Point to)
        {
            bmp.Lock();
            // Get a bounding box for the painted area
            BoundingBox bitmapbounds = new BoundingBox();
            BoundingBox segmentbounds = new BoundingBox();

            //add the size of bmp to the bounding box
            bitmapbounds.AddPoint(0, 0, 0);
            bitmapbounds.AddPoint(bmp.PixelWidth - 1, bmp.PixelHeight - 1, 0);

            UInt32* start = (UInt32*)bmp.BackBuffer.ToPointer();
            int stride = bmp.BackBufferStride / sizeof(UInt32);
            int toX = (int)to.X;
            int toY = (int)to.Y;

            //get the texture image
            var textureImage = backImage1.WriteableBitmap;

            //get the template sizes
            int width = (int)circleImage1.WriteableBitmap.PixelWidth;
            int height = (int)(circleImage1.WriteableBitmap.PixelHeight);

            //add the bounderis of the painted location
            segmentbounds.AddPoint((int)to.X - width / 2, (int)to.Y - height / 2, 0);
            segmentbounds.AddPoint((int)to.X + width / 2 +width%2, (int)to.Y + height / 2 + height%2, 0);
            segmentbounds.Clip(bitmapbounds);

            // for a random location on the texture image
            Random rnd = new Random();

            //add pixel by pixel the ink image to the bmp 
            for (int k = segmentbounds.Top, row = rnd.Next((int)textureImage.PixelHeight - height), bY = 0; k < segmentbounds.Bottom; bY++, k++, row++)
                for (int j = segmentbounds.Left, col = rnd.Next((int)textureImage.PixelWidth - width), bX = 0; j < segmentbounds.Right; bX++, j++, col++)
                {
                    //take the texture image pixel color
                    Color color = textureImage.GetPixel(col, row);

                    //take the template image pixel color
                    Color template = circleImage1.WriteableBitmap.GetPixel(bX, bY);
                    WriteAlphaBlended(start + stride * k + j, Color.FromArgb(template.A, color.R, color.G, color.B));
                }

            //update the bmp
            bmp.AddDirtyRect(new Int32Rect(segmentbounds.Left, segmentbounds.Top, segmentbounds.Width, segmentbounds.Height));
            bmp.Unlock(); 
        }

        // Alpha blends a color with its destination pixel using the standard formula
        private static unsafe void WriteAlphaBlended(UInt32* pixel, Color c)
        {
            byte* component = (byte*)pixel;
            ushort alpha = (ushort)c.A;
            component[3] += (byte)(((ushort)(255 - component[3]) * alpha) / 255);
            component[2] = (byte)(((ushort)component[2] * (255 - alpha) + (ushort)c.R * alpha) / 255);
            component[1] = (byte)(((ushort)component[1] * (255 - alpha) + (ushort)c.G * alpha) / 255);
            component[0] = (byte)(((ushort)component[0] * (255 - alpha) + (ushort)c.B * alpha) / 255);
        }

        public static WriteableBitmap Resize(WriteableBitmap wBitmap, double reqWidth, double reqHeight)
        {
            TransformedBitmap tbBitmap = new TransformedBitmap(wBitmap,
                                             new ScaleTransform(reqWidth / wBitmap.PixelWidth,
                                                                reqHeight / wBitmap.PixelHeight,
                                                                0, 0));
            return BitmapFactory.ConvertToPbgra32Format(tbBitmap);
        }

        private class MovingBrushStroke
        {
            private double xVelocity;
            private double yVelocity;
            private Color color;
            private double lastX;
            private double lastY;
            public MovingBrushStroke(Point from, Point to, Color color)
            {
                TimeToLive = 5;
                xVelocity =  to.X - from.X;
                yVelocity =  to.Y - from.Y;
                lastX = from.X;
                lastY = from.Y;
                this.color = color;
            }
            public int TimeToLive {get; private set;}
            public void DrawNext(WriteableBitmap bmp)
            {
                Random random = new Random();
 
                lastX += xVelocity/2;
                lastY += yVelocity/2;
                double ellipseXCenter = lastX + (random.NextDouble() - 0.5)*10;
                double ellipseYCenter = lastY + (random.NextDouble() - 0.5)*10;
                double radius = random.NextDouble() * 10;
                bmp.FillEllipseCentered((int)ellipseXCenter, (int)ellipseYCenter, (int)radius,(int)radius,color);
                TimeToLive -= 1;
            }


        }
    }


}
