﻿using KinectInteraction;
using Microsoft.Kinect;
using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Mail;
using System.Text;
using System.Threading.Tasks;
using System.Timers;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;

namespace DrawingBoard
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        /// <summary>
        /// Drawing image that we will display the skeleton
        /// </summary>
        private WriteableBitmap imageSource;

        /// <summary>
        /// Interact with the kinect
        /// </summary>
        private KinectHelper kinectHelper;

        /// <summary>
        /// Selected brush
        /// </summary>
        private Brush brush;

        /// <summary>
        /// Edit the image source
        /// </summary>
        private BitmapHelper bitmapHelpers;

        /// <summary>
        /// current color
        /// </summary>
        private Color color = Colors.Black;

        public MainWindow()
        {
            //Open starting window
            var startingWindow = new StartingWindow();
            startingWindow.Show();
            System.Threading.Thread.Sleep(3000);

            //init KinectHelper
            kinectHelper = new KinectHelper();

            
            if (kinectHelper.HandsTracer == null || kinectHelper.KinectImageHelper == null)
            {
                startingWindow.Close();
                MessageBox.Show("The application failed to connect with the Kinect Sensor.\nThe application is shutting down.");
                Application.Current.Shutdown();
                return;
            }

            //init image source
            imageSource = new WriteableBitmap(kinectHelper.KinectImageHelper.FrameWidth,
                kinectHelper.KinectImageHelper.FrameHeight, 96.0, 96.0, PixelFormats.Pbgra32, null);

            //set air brush to default brush
            brush = Brush.AirBrush;

            //init bitmapHelpers
            bitmapHelpers = new BitmapHelper();

            InitializeComponent();

            //resigter on the HandsLocationChanged event with the function HandsTracer_HandsLocationChanged
            kinectHelper.HandsTracer.HandsLocationChanged += HandsTracer_HandsLocationChanged;

            //close starting window
            startingWindow.Close();
        }

        /*
        /// <summary>
        /// Set the next color
        /// </summary>
        private void ChangeColor(object sender, ElapsedEventArgs e)
        {
            color++;
            if (color > colors.Count-1)
                color = 0;
        }*/

        /// <summary>
        /// set Image1 source to imageSource
        /// </summary>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {  
            Image1.Source = imageSource;
        }

        /// <summary>
        /// 
        /// </summary>
        void HandsTracer_HandsLocationChanged()
        {
            DrawHand( kinectHelper.HandsTracer.LeftHand);
            DrawHand( kinectHelper.HandsTracer.RightHand);
        }

        /// <summary>
        /// Draw an hand 
        /// </summary>
        /// <param name="hand"></param>
        private void DrawHand( Hand hand)
        {
            //if hand not Tracked do nothing
            if(hand.State != JointTrackingState.Tracked || (hand.OldPosition.X == 0 && hand.OldPosition.Y == 0))
                return;

            switch (brush)
            {
                case Brush.AirBrush:
                    bitmapHelpers.Airbrush(imageSource,
                        PointToScreen(hand.OldPosition), PointToScreen(hand.Position),
                        color,
                        6 * 2);
                    break;
                case Brush.Ellipse:
                    bitmapHelpers.Ellipse(imageSource,
                        PointToScreen(hand.OldPosition), PointToScreen(hand.Position),
                        color);
                    break;
                case Brush.Stroke:
                    bitmapHelpers.MovingBrush(imageSource,
                        PointToScreen(hand.OldPosition), PointToScreen(hand.Position),
                        color);
                    break;
                case Brush.ImageDraw:
                    bitmapHelpers.DrawImage(imageSource, PointToScreen(hand.Position),color);
                    break;
                case Brush.BackImageDraw:
                    bitmapHelpers.DrawBackImage(imageSource, PointToScreen(hand.Position));
                    break;
                default:
                    return;
            }   
        }

        /// <summary>
        /// convert from kinect resolusion to screen resolution.
        /// </summary>
        /// <param name="point"></param>
        /// <returns></returns>
        private Point PointToScreen(SkeletonPoint point)
        {
            // Convert point to depth space.  
            // We are not using depth directly, but we do want the points in our 640x480 output resolution.
            DepthImagePoint depthPoint = kinectHelper.MapPointToDepthPoint(point, DepthImageFormat.Resolution640x480Fps30);
            return new Point(depthPoint.X, depthPoint.Y);
        }

        void SelectAirBrush(object sender, RoutedEventArgs e)
        {
            brush = Brush.AirBrush;
        }

        private void SelectEllipse(object sender, RoutedEventArgs e)
        {
            brush = Brush.Ellipse;
        }

        private void SelectStroke(object sender, RoutedEventArgs e)
        {
            brush = Brush.Stroke;
        }

        private void SelectImageDraw(object sender, RoutedEventArgs e)
        {
            brush = Brush.ImageDraw;
        }

        private void SelectBackImageDraw(object sender, RoutedEventArgs e)
        {
            brush = Brush.BackImageDraw;
        }

        /// <summary>
        /// Load new back image
        /// </summary>
        private void LoadImage(object sender, RoutedEventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            // Set filter for file extension and default file extension
            dlg.Filter = "Image files (*.jpg, *.jpeg, *.jpe, *.jfif, *.png) | *.jpg; *.jpeg; *.jpe; *.jfif; *.png";
            // Display OpenFileDialog by calling ShowDialog method
            bool? result = dlg.ShowDialog();
            // Get the selected file name and display in a TextBox
            if (result == true)
            {
                string filename = dlg.FileName;
                bitmapHelpers.LoadNewBackImage(filename);
                BackImage.Source = new BitmapImage(new Uri(filename));
            }
        }


        private void SendMail(object sender, RoutedEventArgs e)
        {
            var mailWindow = new MailWindow();
            mailWindow.ShowDialog();
            if (string.IsNullOrEmpty(mailWindow.MailAdress))
                return;

            //Holds message information.
            MailMessage mailMessage = new MailMessage();
            //Add basic information.
            mailMessage.From = new MailAddress("kinectcstechnion@gmail.com");
            mailMessage.To.Add(mailWindow.MailAdress.Trim());

            mailMessage.Subject = "Your image";
            CreateFileFromBitmapSource(imageSource.Clone());
            mailMessage.Attachments.Add(new Attachment(Path.GetTempPath() + "Image.png"));
            var smtpClient = new SmtpClient
            {
                Host = "smtp.gmail.com",
                Port = 587,
                EnableSsl = true,
                DeliveryMethod = SmtpDeliveryMethod.Network,
                UseDefaultCredentials = false,
                Credentials = new NetworkCredential("kinectcstechnion@gmail.com", "cstechnion")
            };
            smtpClient.Send(mailMessage);
        }

        //create image file from bitmap 
        void CreateFileFromBitmapSource(BitmapSource image)
        {
            using (FileStream stream = new FileStream(Path.GetTempPath() + "Image.png", FileMode.Create))
            {
                PngBitmapEncoder encoder = new PngBitmapEncoder();
                encoder.Frames.Add(BitmapFrame.Create(image));
                encoder.Save(stream);
                stream.Close();
            }
        }

        //for debug
        private void Image1_MouseDown_1(object sender, MouseButtonEventArgs e)
        {
            DrawHand(kinectHelper.HandsTracer.LeftHand);
        }

        //Select new color
        private void ColorPicker_SelectedColorChanged(object sender, RoutedPropertyChangedEventArgs<Color> e)
        {
            color = e.NewValue;
        }

        private void Slider_ValueChanged_1(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if (kinectHelper.HandsTracer == null || kinectHelper.KinectImageHelper == null)
                return;
            if(e.NewValue >10 && e.NewValue <100)
                bitmapHelpers.SetDrawImageSize((int)e.NewValue);
        }
    }

    enum Brush { AirBrush, Ellipse, Stroke, ImageDraw, BackImageDraw }
}
